;ident:8,24
;EDITOR part of SID-Wizard
;=====================================CONSTANTS===========================================
;CAUTION: SEQUENCE/PATTERN/TABLE DELIMITERS ($FE/$FF/$7E/$7F,TEMPOTABLE BIT7) ARE HARDWIRED, NEVER EVER TRY TO MODIFY THEM!!!
; $Id: editor.asm 166 2012-11-07 22:25:17Z hermitsoft $

COMPILEDAPP = 1 ;1 stands for the 'editor'

		.include "settings.cfg";global constants to fine-tune the maximal memory occupation of parts of tune data

;====================================== VOLATILE DATA RANGE ===========================
;don't use .byte or .fill, because it will alter the start-address of the compiled program
zerovar= $0340			;here are the variables that are used by the program, and initially must be set zero.
						;They remain till program executes. (Displayer needs all these variables to be here for faster batch-indexing operations.)
actptn1=zerovar+0		;.byte $00 ;selected (or played in follow-mode) pattern 1
actptn2=zerovar+1		;.byte $00 ;selected (or played in follow-mode) pattern 2
actptn3=zerovar+2		;.byte $00 ;selected (or played in follow-mode) pattern 3

zerovar2=zerovar+3
actwind=zerovar2+0 		;.byte 0  ;actual window (0-patterneditor,1-orderlist,2-instrum,3-chord,4-tempo)

prowpos=zerovar2+1 		;.byte $0,$0,$0  ;first display-row positions of the patterns (track1..3)
ptncurs=zerovar2+4 		;.byte 0,0,0 ;memory index in patterns under cursor Y-position (displayer1 calculates) - in note column only

ptn1pos=zerovar2+7		;.byte $00 ;pattern1 display start-position in memory - calculated and used by displayer1
ptn2pos=zerovar2+8		;.byte $00 ;pattern2 display start-position in memory - calculated and used by displayer1
ptn4pos=zerovar2+9		;.byte $00 ;pattern3 display start-position in memory - calculated and used by displayer1

ptrdysi=zerovar2+10		;.byte $00 ;signal that notifies key-repeater to wait for pattern-redrawing

subwpos1=zerovar2+11	;.byte $0,$0,$0,$0,$0 ;subwindow/position 1 in actual window
subwpos2=zerovar2+16	;.byte $0,$0,$0,$0,$0 ;subwindow/position 2 in actual window
subwpos3=zerovar2+21	;.byte $0,$0,$0,$0,$0 ;subwindow/position 3 in actual window
inswXbuf=zerovar2+26	;.byte $0,$0,$0,$0 ;instrument-subwindow X positions
inswYbuf=zerovar2+30	;.byte $0,$0,$0,$0 ;instrument-subwindow Y positions

;olstart=zerovar2+34		;.byte $00,$00,$00 ;orderlist F2key-playing startpositions

;these 5 bytes are index-handled some times, so leave them like this
wfarpos=zerovar2+37		;.byte 0 ;wfarp-table display-poisition
pwtbpos=zerovar2+38		;.byte 0 ;pwtable display-position
ctftpos=zerovar2+39		;.byte 0 ;ctftable display-position
chorpos=zerovar2+40		;.byte 0 ;chordtable display-position
temppos=zerovar2+41		;.byte 0 ;tempotable display-position
;---------------

songtime=zerovar2+42	;.word $0000  ;1st byte - minutes in BCD , 2nd byte - seconds in BCD
timeroff=zerovar2+44	;.byte 0  ;timer ON/OFF switch

namemode=zerovar2+45	;.byte 0  ;if value is nonzero, keyboard routine is typing characters instead of inst./pattern/etc. editing
						   ;if value is 2, filename-typing/selection mode
menumode=zerovar2+46	;.byte 0  ;if value is nonzero, the menu appears above music editor (display of musicdata stops)

dovetail=zerovar2+47	;.byte 0  ;if 1, cycling tracks on note entry/jam

OSC3Buffer=zerovar2+48	;.fill 12		;storage for OSC3 SID output to display oscilloscope
						;ENV3Buffer .fill 12 ;storage for ENV3 SID output to display oscilloscope

;------------------------------------------Video RAM is used for upper border, disappears when opening loader/saver/dir.display
lowsprites=$0400		;.align $100
ordlspr=lowsprites+0	; .fill $200		;orderlist-spritedata, and other volatile data can be here
orlcurs=lowsprites+$200	; .fill $40		;orderlist-cursor sprite - copied here from RAM unde I/O
orlcur2=lowsprites+$240	; .fill $40		;alternative orderlist-sprite for NTSC mode (due to sprite-coordinate issues)
orlcur3=lowsprites+$280	; .fill $40		;alternative orderlist-sprite for NTSC mode (due to sprite-coordinate issues)

seqbuffer=lowsprites+$2c0		;around 128 byte buffer where
		;.byte $fe		;initial empty sequence-region - a delimiter
		;.fill seqbound-1,0


.cerror *>$0800,"TOO MUCH DATA BEFORE EDITOR DOESN'T FIT IN MEMORY. MOVE SOME DATA TO UPPER FREE AREAS!"
;================================= BASIC area start ===========================
		*= $0801
		.word ss,Year
		.null $9e,^start ;Sys 2064
ss		.word 0

		*= $080d		;SYS 2061 to safe-restart the editor, without initializing patterns/instruments
		jmp SafeRestart
;================================= Code part 1 ================================
		*= $0810		;INITIALIZATION AT PROGRAM START - set IRQ handlers, screen, and main VIC registers
start	sei
		lda #$35
		sta banksel
		jsr STARTUPMENU.doit ;detect PAL/NTSC, select player, etc. - volatile, overwritten by patterns at next step
		jsr menu.CINITUN		;clear & initialize music data area in menu.inc
		jsr menu.CININST		;clear & initialize instrument-data area

;------------------------------------------------------------------------------
started sei				;we don't want lost cycles by IRQ calls :)
appiniter				;(this section was contained by initer.inc before)
		lda #$34
		sta banksel
		ldy #0
-		lda #0
		sta zerovar2,y	;zerovar,y
		sta ordlspr,y	;zero out orderlist sprite-bitmap area (the gaps between orderlist-rows won't be cleared afterwards)
		sta ordlspr+$100,y
		sta orlcurs,y	;zero sprites to initialize not written parts
		iny
		bne - ;Y=0 on exit
		sta insdelcopy.seqbufs+1 ;init sequence-buffer (set sequence-buffer size to zero)

		ldy #$3E
-		lda olcursp,y ;copy orderlist-cursor sprite from RAM under IO area
		sta orlcurs+0,y ;and overwrite the beginning of orlcur2 with zeroes
		sta orlcur2+(1*NTSCseqdistance)*3,y ;copy alternative shifted orderlist-cursor for NTSC mode
		sta orlcur3+(2*NTSCseqdistance)*3,y ;copy alternative shifted orderlist-cursor for NTSC mode
		dey
		bpl -

		inc banksel ;$35 ;no need for ROM, just when loading/saving in menu
		lda #$7f
		sta $dc0d
		sta $dd0d
		lda $dc0d
		lda $dd0d
		lda #<irq
		sta $fffe		;maskable IRQ pointer setting
		lda #>irq
		sta $ffff
		lda #<dummirq
		sta $fffa		;NMI (restore key) pointer setting
		lda #>dummirq
		sta $fffb

		lda ThemeNo
		jsr coloriz		;based on Accumulator, select a colour-theme
		jsr iniscrn		;initialize VIC, screen colors and sprite-pointers

		;$01 must be $35 here

		lda #(cursprite//$4000 / $40);modulo of VIC-banksize divided by spritesize
		sta $fff8		;cursor-sprite pointer for 26th row
		sta vidram+$3f8

		ldx #1
		stx $d01a		;enable raster-IRQ
		stx $d012		;(IRQ will wait till row $F4 anyway)
		;stx inirequ+1 ;ensure that player initialization happens at startup

		jsr playadapter.resetune	;contains inireq+1 increasing + selects 1st patterns of subtune
		cli


;---------------------- display-refresher main loop (outside IRQ) ------------------
mainloop
inirequ lda #1			;self-written variable ;if not 0, signs tune initialization request for main routine
		beq +			;check init request from IRQ
initune jsr playadapter.inisubb		;player-adapter version of initer ($1000) of player routine
		;jsr ptndisp		;display selected patterns
		;jsr stpdisp		;display step-highlighted bars for track 1..3
		lda #0
		sta inirequ+1
+		jsr display		;considers menu-display over patterns (based on 'menumode' variable)
		lda menu.menupoint	;check selected menupoint (value is 0 if no menupoint was launched)
		beq mainloop	;if no menu was launched, main loop continues
		jmp menu.runmenup

SafeRestart
		sei
		lda #$35
		sta banksel
		ldy #$fe
		;sty inirequ+1 ;ensure that player initialization happens at startup
		sty $dc00
		iny				;wait for release of RETURN-key
		cpy $dc01
		bne *-3
		jmp menu.retoedi		;started
;==============================================================================
		.include "include/irq.inc";call keyhandler, cursordisplay, player routine, raster-tricks, followplay
 ;(don't move this area if possible, because code boundary-crossing can affect rastertiming)
;==============================================================================
		.include "include/displayer1.inc";display pattern, step highlighting & numbering, etc.

;==============================================================================
;							  end of code part 1
;==============================================================================
		.cerror *>PLAYERADDR,"TOO MUCH CODE BEFORE PLAYER DOESN'T FIT IN MEMORY. MOVE SOME CODE TO UPPER FREE AREAS!"

;************************ MUSICAL CODE AND DATA *******************************
		*= PLAYERADDR	;$1000
MAX_PLAYERSIZE=$0A00
player .block
 .include "include/player.asm";the player/driver routine to produce the SIDDY sound
.bend

;-------------------------------
.cerror *>(PLAYERADDR+MAX_PLAYERSIZE),"!!! PLAYER IS TOO BIG !!! OPTIMIZE IT OR USE IN ZEROPAGE-VARIABLE MODE !!!"
*=PLAYERADDR+MAX_PLAYERSIZE	;give enough space

		#FeatInstances ;macro in settings.cfg

		.include "include/playadapter.inc";subroutines connecting player and editor (jamming, follow-play, etc.)


;===============================
MUSICDATA				;musicdata-pointers reset relative to this address accordingly...preset music-data for editor------------------
;---------------------FIX pointer-tables-no need to save and recalculate on load-----------------------
		*= SUBTUNES		;Subtune-orderlist-pointer table (max. 256 bytes, packer will crop unused areas)
		.for SUBTC=0,SUBTC<=maxsubtuneamount,SUBTC=SUBTC+1
		.byte <(SEQUENCES+SUBTC*3*seqbound),>(SEQUENCES+SUBTC*3*seqbound)
		.byte <(SEQUENCES+SUBTC*3*seqbound+seqbound),>(SEQUENCES+SUBTC*3*seqbound+seqbound)
		.byte <(SEQUENCES+SUBTC*3*seqbound+seqbound*2),>(SEQUENCES+SUBTC*3*seqbound+seqbound*2)
		.byte defsubtempo1,defsubtempo2;DEFAULT FUNKSPEED FOR SUBTUNES
		.next
		*= PPTRLO		;Pattern lo-pointer-table (max. 128 bytes, packer will crop unused areas)
		.byte <PTN0
		.for PTPC=0,PTPC<maxptnamount,PTPC=PTPC+1;LO-BYTES OF PATTERN-POINTERS
		.byte <(RESTPTN+PTPC*PTNBOUND)
		.next
		*= PPTRHI		;Pattern hi-pointer-table (max. 128 bytes, packer will crop unused areas)
		.byte >PTN0
		.for PTPC=0,PTPC<maxptnamount,PTPC=PTPC+1;HI-BYTES OF PATTERN-POINTERS
		.byte >(RESTPTN+PTPC*PTNBOUND)
		.next
		*= INSPTLO+1	;Instrument-data lo-Pointer Table ;(max. 128 bytes, packer will crop unused areas), instrument 0 unused
		.for INSCL=0,INSCL<maxinstamount,INSCL=INSCL+1;LO-BYTES OF INSTRUMENT-POINTERS
		.byte <(INSTRUMENTS+INSCL*maxinstsize)
		.next
		*= INSPTHI+1	;Instrument-data hi-Pointer Table ;(max. 128 bytes, packer will crop unused areas)
		.for INSCH=0,INSCH<maxinstamount,INSCH=INSCH+1;HI-BYTES OF INSTRUMENT-POINTERS
		.byte >(INSTRUMENTS+INSCH*maxinstsize)
		.next
;---------------------variable-pointer tables - not saved but need to be recalculated---------------------
		*= CHDPTRLO		;Chord-pointer Table (max. 64 bytes, packer will crop unused areas)
		.byte 0,1		;unchanged chord-pointer of chord 0 and 1
		*= TEMPTRLO		;Pointers to specific tempo-programs in tempo-program table (max. 128 bytes, packer will crop unused areas)
		.byte 8,9
;...ptnlength - pattern length registry for faster operations and display  - JUST FOR EDITOR ;(max. 128 bytes, packer will crop unused areas)
;...ptnsize - real size of pattern in memory - JUST FOR EDITOR
;=====================music data to save starts here - in the same order as saved file==============================
		*= TUNEHEADER	;some basic tune-specific settings workfile/tune/subtune related settings (framespeed, etc.)
filetyp .text "SWM"		;3 BYTE FILETYPE-DESCRIPTION - abbreviation of 'SID-WIZARD MODULE'
version .text SWMversion ; VERSION 1 of the module format
framesp .byte 1			;framespeed of the music (1 singlespeed.. to 8 multispeed)
stpthck .byte $04		;thick amount for step-highlighting
autoadv .byte 1			;switch for autoadvance ON/OFF - saved with tune
confbit .byte %00000001	;ON/OFF configurations, each bit corresponds to a particular SID-Wizard ON/OFF-setting
 ;bit0 (1) - bind tracks ON/OFF (ON=1) - move track together wherever possible, or not at all
 ;bit1 (2) - rasterbar-display ON/OFF ('ON=0' to keep compatibility with previous SWM files) (some people get annoyed by player-rasterbars)
 ;bit2 (4) - followplay as default?  ('ON=1' to keep compatibility) - toggle follow-play for simple F1 play
 ;bit7 ($80) - instrument auto-append ON/OFF ('ON=1' to keep compatibility) - whether instrument should be typed with note automatically
mutesolo .byte $FF,$FF,$FF;the mute/solo indicators for the tracks (toggled by shift+1/2/3)
defptlen .byte DEFAULTPTNLEN;default pattern-length (this info is needed for depacker to retain the status before saving)
;.............CALCULATED VALUES FOR PACKER/DEPACKER follow here (described in SWM-spec.src)
.fill 5 ;bytes modified by packer/depacker (their positions are given in settings.cfg)
ThemeNo	.byte 0 ;colour-theme number
KeybTyp	.byte 0 ;keyboard-layout type  (0:default SID-Wizard, 4:FC/GMC, etc.)
DrvType .byte 0 ;<(PLAYERID) ;type of the music player/driver routine (0=FULL, 1=MID, 2=LIGHT, 3=EXTRA), just an information, no restriction caused
;...some bytes left for later additions (to keep compatibility)
		*= authorinfo
		.text "AUTHOR NAME : TUNE-TITLE...             "
;...SEQUENCES - orderlists of subtunes ;maximum $2000 bytes ( $2000/3tracks*$80 = 20 SUBTUNES )
;...PATTERNS ;maximum $8000 bytes	($7f00/$100 = $7f patterns)
		*= PTN0			;dummy pattern-data (in case f empty subtune-playback)
		.byte $00,$ff

		;=======================================================================
		.include "include/startupmenu.inc" ;initially at pattern-area, then overwritten by tune-initer
        ;contains player-types (normal/medium/light/extra), one-shot help, and a menu where everything can be selected (player/PAL-NTSC/etc.)
		;=======================================================================

;...INSTRUMENTS ;maximum $1000 bytes. $1000/$3e = $42 (~64) bytes for each, but there
		*= CHORDS		;Cordtable - dynamic memory management with aid of pointers (max. 256 bytes, packer will crop unused areas)
		.byte $7f		;- ensure delimiter as chord0
;...TEMPOTBL, TRKTEMPO, RESTEMP - tempo program-table (max. 256 bytes, packer will crop unused areas)
		*= RESTEMP
		.byte $80		;1st tempo if accidentally tempo-program 0 (at index 8) is selected
		;and the rest of tempoprograms coming here
		*= endoftune


;================================= COPY/PASTE buffers =============================================
patternbuffer			;around 256 byte buffer where the patterns are being copied
		.byte 0,$ff		;initial value (empty pattern)
		.fill PTNBOUND-2,0;zero out copybuffer area

instbuffer ;fill instrument copy-buffer with default values
defIbas	.byte %00011010	;0 CONTROL REGISTER (HR TIMER,HRTYPE ,VIBRATO-TYPE,TIED PW/CTF)
		.byte >DEFAULTHRADSR,<DEFAULTHRADSR;1 HR-ADSR
		.byte $00,$F4	;3 ADSR ON NOTESTART (GATE-ON)
		.byte $84		;5 VIBRATO (CALCULATED) FREQ+AMP (amplitude:high nibble)
		.byte $10		;6 VIBRATO DELAY OR AMPLITUDE-INCREMENT SPEED
		.byte $00		;7 ARPEGGIO AND CHORD-SPEED (0=1X)
		.byte $01		;8 DEFAULT CHORD FOR INSTRUMENT
		.byte $00		;9 OCTAVE SHIFT (2's complement) - it's transpose essentially but in 12 steps by editor
		.byte <(defPWba - defIbas) ;PWTBL1-INST1 ;A PW-TABLE POINTER RELATIVE TO INSTRUMENT-ADDRESS
		.byte <(defFIba - defIbas)	;FLTBL1-INST1 ;B FILTER-TABLE POINTER RELATIVE TO INSTRUMENT-ADDRESS
		.byte $00		;C GATEOFF POINTER FOR WFARP-TABLE RELATIVE TO WF-TABLE POSITION
		.byte $00		;D GATEOFF POINTER FOR PW-TABLE RELATIVE TO PW-TABLE POSITION
		.byte $00		;E GATEOFF POINTER FOR FILTER-TABLE RELATIVE TO FILTER-TABLE POS.
		.byte $09		;F 1ST FRAME WAVEFORM
		.byte $21,0,0,$ff ;WF-PROGAM ;hardwired position at $10 (relative to instrument-base)
defPWba	.byte $ff		;PW-PROGRAM
defFIba	.byte $ff		;FILTER-PROGRAM
		*=instbuffer+(maxinstsize-instnamelength)
		.text "SAW-SOLO"

;-------------------------code that had no room elsewhere---------------------------------------
;------------------------------pack/depack/(load/save if no BASIC-ROM needed)---------------------------
		.include "include/commonsubs.inc";common/miscellaneous subroutines used by various parts of the program. pointer/index calculators, timer, etc.
		.include "include/menu.inc";menu and file-dialog and file-operations
		.include "include/packdepack.inc";compresses/decompresses the tune to prepare for loading/saving by wiping out empty area

		.dsection data	;tables/variables being displayed and used ;all data from other sources (between '.section data' and '.send data' comes here
		.send data		;(was 'datatables.inc' originally but it went obsolete as now all data can be placed next to its code)

;************************ CHARACTER SETS AND SCREEN DATA **********************
		.cerror *>$D000,"TOO MUCH MUSICDATA DOESN'T FIT IN MEMORY. LESSEN PATTERNS OR INSTRUMENTS OR SUBTUNES!"
;==============================================================================
		*= $d000		;screendata - videoram and colour-ram initial GUI design, sprite pointers,sprites
		.include "graphics/graphics.inc"
;$d000 - c64 character set for SID-Wizard (binary file included)
;$d800 - extra RAM under IO used for non-graphic, or grahpic-source content
 ;seqchtbl ;reference table (5X multiplier) for 5-pixel high chardata
 ;sprtable ;precalc.table for sprite-addresses of corresponding orderlist x-coordinates
 ;seqchar	;characters for orderlist patternvalues' display
 ;cursor-sprite in orderlist
 ;cursor-sprite
;around $dc00: Code in RAM Under IO area (section CodeUnderIO)
;$e000..$e400 VIDEO RAM initial and maintained content
;==============================================================================

;********************************** CODE PART 2 *******************************
*=$e400

;IO-bank switchers for insdelcopy.inc public procedures - needed because insdelcopy.inc was moved to RAM under IO
CopyIns dec banksel ;$34
		jsr insdelcopy.CopyIns
		jmp returIO
PasteIn dec banksel ;$34
		jsr insdelcopy.PasteIn
		jmp returIO
copytobuf dec banksel ;$34
		jsr insdelcopy.copytobuf
		jmp returIO
transptn dec banksel ;$34
		jsr insdelcopy.transptn
		jmp returIO
cutandcopy dec banksel ;$34
		jsr insdelcopy.cutandcopy
		jmp returIO
pastebuf dec banksel ;$34
		jsr insdelcopy.pastebuf
		jmp returIO
delbackpt dec banksel ;$34
		jsr insdelcopy.delbackpt
		jmp returIO
insertpt dec banksel ;$34
		jsr insdelcopy.insertpt
		jmp returIO
copyseq	dec banksel ;$34
		jsr insdelcopy.copyseq
		jmp returIO
pasteseq dec banksel ;$34
		jsr insdelcopy.pasteseq
returIO	inc banksel ;$35
		rts
		.section CodeUnderIO  ;in graphics.inc area $dc00...$dfff
		.include "include/insdelcopy.inc";batch-manipulators, like insertion/deletion/cut/copy/paste/transpose/etc. 
		.send CodeUnderIO

;----------------------------------------------------------------------------------
		.include "include/keyhandler.inc";handle the keypresses - the subroutine-names are found in the datatables.inc file ;callback subroutines for keypresses
		.include "include/datawriter.inc";common point to write value in X to cursor-position, and note in Y to cursor position
;==============================================================================
		.include "include/displayer2.inc";display cursor and textual content (mainly orderlist)

;...not too much bytes left here
		.cerror (*>$ffc0 || *<$e400),"TOO MUCH CODE DOESN'T FIT IN MEMORY. MOVE SOME CODE TO LOWER FREE AREAS!"
;==============================================================================
;$ffc0 ;row26  ;videoram for 26th row
		*= $ffc0		;videoram for 26th row
row26

;===================================================================================================================================
; vim: sw=4 ts=4 syntax=asm:
